@@ -24,17 +24,42 @@ namespace client
 {
 
 	I2CPDestination::I2CPDestination (std::shared_ptr<I2CPSession> owner, std::shared_ptr<const i2p::data::IdentityEx> identity, bool isPublic, const std::map<std::string, std::string>& params):
-		LeaseSetDestination (isPublic, &params), m_Owner (owner), m_Identity (identity)
+		RunnableService ("I2CP"), LeaseSetDestination (GetIOService (), isPublic, &params), 
+		m_Owner (owner), m_Identity (identity), m_EncryptionKeyType (m_Identity->GetCryptoKeyType ())
 	{
 	}
 
+	I2CPDestination::~I2CPDestination ()
+	{
+		if (IsRunning ())
+			Stop ();
+	}	
+		
+	void I2CPDestination::Start ()
+	{
+		if (!IsRunning ())
+		{	
+			LeaseSetDestination::Start ();
+			StartIOService ();
+		}	
+	}
+		
+	void I2CPDestination::Stop ()
+	{
+		if (IsRunning ())
+		{	
+			LeaseSetDestination::Stop ();
+			StopIOService ();
+		}	
+	}	
+		
 	void I2CPDestination::SetEncryptionPrivateKey (const uint8_t * key)
 	{
 		memcpy (m_EncryptionPrivateKey, key, 256);
 		m_Decryptor = i2p::data::PrivateKeys::CreateDecryptor (m_Identity->GetCryptoKeyType (), m_EncryptionPrivateKey);
 	}
 
-	bool I2CPDestination::Decrypt (const uint8_t * encrypted, uint8_t * data, BN_CTX * ctx) const
+	bool I2CPDestination::Decrypt (const uint8_t * encrypted, uint8_t * data, BN_CTX * ctx, i2p::data::CryptoKeyType preferredCrypto) const
 	{
 		if (m_Decryptor)
 			return m_Decryptor->Decrypt (encrypted, data, ctx, true);
@@ -221,8 +246,16 @@ namespace client
 			m_PayloadLen = bufbe32toh (m_Header + I2CP_HEADER_LENGTH_OFFSET);
 			if (m_PayloadLen > 0)
 			{
-				m_Payload = new uint8_t[m_PayloadLen];
-				ReceivePayload ();
+				if (m_PayloadLen <= I2CP_MAX_MESSAGE_LENGTH)
+				{	
+					m_Payload = new uint8_t[m_PayloadLen];
+					ReceivePayload ();
+				}
+				else
+				{
+					LogPrint (eLogError, "I2CP: Unexpected payload length ", m_PayloadLen); 
+					Terminate ();
+				}	
 			}
 			else // no following payload
 			{
@@ -547,7 +580,7 @@ namespace client
 					uint16_t keyType = bufbe16toh (buf + offset); offset += 2; // encryption type
 					uint16_t keyLen = bufbe16toh (buf + offset); offset += 2;  // private key length
 					if (offset + keyLen > len) return;
-					if (keyType > currentKeyType)
+					if (!currentKey || keyType > currentKeyType)
 					{
 						currentKeyType = keyType;
 						currentKey = buf + offset;
@@ -556,7 +589,10 @@ namespace client
 				}				
 				// TODO: support multiple keys
 				if (currentKey)
+				{
 					m_Destination->SetEncryptionPrivateKey (currentKey);
+					m_Destination->SetEncryptionType (currentKeyType);
+				}
 
 				m_Destination->LeaseSet2Created (storeType, ls.GetBuffer (), ls.GetBufferLen ()); 
 			}
@@ -786,8 +822,11 @@ namespace client
 	{
 		m_IsRunning = false;
 		m_Acceptor.cancel ();
-		for (auto& it: m_Sessions)
-			it.second->Stop ();
+		{
+			auto sessions = m_Sessions; 
+			for (auto& it: sessions)
+				it.second->Stop ();
+		}	
 		m_Sessions.clear ();
 		m_Service.stop ();
 		if (m_Thread)
